// play.cpp: conceptual audio/video stream decoding example
// Notes: 08/30/2015 GL based on play.sc
//        09/04/2015 GL use 'long long' (not 'double') to avoid floating-point overflow
//                      (FPU exception handling shows high performance variation;
//                      moreover, integer arithmetic is more realistic for decoder)
// Author: Guantao Liu, CECS
// Date: 08/30/2015
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "systemc.h"
// frame rates
#define AUDIO_DELAY	sc_time(26.12, SC_MS)
#define VIDEO_DELAY	sc_time(33.33, SC_MS)
// patched frame rates
#define AUDIO_DELAY_P	26120, SC_US
#define VIDEO_DELAY_P	33330, SC_US
// simulated time
#define CLIP_LENGTH	sc_time(10, SC_SEC)
// simulated workload unit
#define AUDIOLOAD	 7500000
#define VIDEOLOAD	10000000
// illustrate frame decoding
//#define PRINT_TIMES
#define PRINT_FRAMES
// queue communication

class i_long_long_sender : virtual public sc_interface
{
  public: virtual void send(long long d) = 0;
}
;

class i_long_long_receiver : virtual public sc_interface
{
  public: virtual void receive(long long *d) = 0;
}
;

class c_long_long_queue : public sc_channel,public i_long_long_sender,public i_long_long_receiver
{
  public: static unsigned int _OoO_Instance_Counter;
  

  inline c_long_long_queue(class sc_module_name name,unsigned long size) : sc_channel(name), m_size(size)
{
    sc_module::m_instance_id = _OoO_Instance_Counter++;
    (this) -> wr = 0;
    (this) -> ws = 0;
    (this) -> p = 0;
    (this) -> n = 0;
    (this) -> buffer = 0L;
  }
  

  inline void setup()
{
    struct chnl_scoped_lock _sc_chnl_scoped_lock(sc_channel::m_mutex);
    if (!((this) -> buffer)) {
      unsigned long i;
      long long dummy;
      if (!((this) -> buffer = ((long long *)(malloc(sizeof(long long ) * (this) -> m_size))))) {
        printf("c_typed_queue\n");
        abort();
      }
      for (i = 0; i < (this) -> m_size; i++) {
        memcpy((&(this) -> buffer[i]),(&dummy),sizeof(long long ));
      }
    }
  }
  

  inline void cleanup()
{
    struct chnl_scoped_lock _sc_chnl_scoped_lock(sc_channel::m_mutex);
    if (!((this) -> n)) {
      free(((this) -> buffer));
      (this) -> buffer = 0L;
    }
  }
  

  virtual inline void receive(long long *d)
{
    struct chnl_scoped_lock _sc_chnl_scoped_lock(sc_channel::m_mutex);
    while(!((this) -> n)){
      (this) -> wr++;
      (this) ->  wait ((this) -> r,0);
      (this) -> wr--;
    }
    if ((this) -> n <= (this) -> p) {
       *d = (this) -> buffer[(this) -> p - (this) -> n];
    }
     else {
       *d = (this) -> buffer[(this) -> p + (this) -> m_size - (this) -> n];
    }
    (this) -> n--;
    if (((this) -> ws)) {
// delta notification
      (this) -> s .  notify (SC_ZERO_TIME);
    }
    (this) ->  cleanup ();
  }
  

  virtual inline void send(long long d)
{
    struct chnl_scoped_lock _sc_chnl_scoped_lock(sc_channel::m_mutex);
    while((this) -> n >= (this) -> m_size){
      (this) -> ws++;
      (this) ->  wait ((this) -> s,0);
      (this) -> ws--;
    }
    (this) ->  setup ();
    (this) -> buffer[(this) -> p] = d;
    (this) -> p++;
    if ((this) -> p >= (this) -> m_size) {
      (this) -> p = 0;
    }
    (this) -> n++;
    if (((this) -> wr)) {
// delta notification
      (this) -> r .  notify (SC_ZERO_TIME);
    }
  }
  private: class sc_event r;
  class sc_event s;
  unsigned long wr;
  unsigned long ws;
  unsigned long p;
  unsigned long n;
  long long *buffer;
  unsigned long m_size;
}
;
unsigned int c_long_long_queue::_OoO_Instance_Counter = 0;

struct Stimulus : public sc_module
{
  static unsigned int _OoO_Instance_Counter;
  sc_core::sc_port< i_long_long_sender  , 1 , sc_core::SC_ONE_OR_MORE_BOUND > VideoStream;
  sc_core::sc_port< i_long_long_sender  , 1 , sc_core::SC_ONE_OR_MORE_BOUND > AudioStreamL;
  sc_core::sc_port< i_long_long_sender  , 1 , sc_core::SC_ONE_OR_MORE_BOUND > AudioStreamR;
  

  inline void main()
{
    class sc_time t1 = sc_time::from_value(0);
    class sc_time t2 = sc_time::from_value(0);
    long long video = 42;
    long long audioLeft = 27;
    long long audioRight = 15;
    while(t1 < sc_time::sc_time(10,sc_core::SC_SEC) || t2 < sc_time::sc_time(10,sc_core::SC_SEC)){
      if (t1 < t2) {
        ((this) -> VideoStream) ->  send (video);
        t1 += sc_time::sc_time(33.33,sc_core::SC_MS);
      }
       else {
        ((this) -> AudioStreamL) ->  send (audioLeft);
        ((this) -> AudioStreamR) ->  send (audioRight);
        t2 += sc_time::sc_time(26.12,sc_core::SC_MS);
      }
    }
    printf("\nDone!\n");
    exit(0);
  }
  typedef struct Stimulus SC_CURRENT_USER_MODULE;
  

  inline Stimulus(class sc_module_name )
{
    sc_module::m_instance_id = _OoO_Instance_Counter++;
{
      class sc_process_handle main_handle = sc_get_curr_simcontext() ->  create_thread_process ("main", false, static_cast<sc_core::SC_ENTRY_FUNC>(&Stimulus::main), (this), 0, 0, sc_module::m_instance_id);
      (this) -> sc_module::sensitive << main_handle;
      (this) -> sc_module::sensitive_pos << main_handle;
      (this) -> sc_module::sensitive_neg << main_handle;
    }
    ;
  }
}
;
unsigned int Stimulus::_OoO_Instance_Counter = 0;

struct VideoCodec : public sc_module
{
  static unsigned int _OoO_Instance_Counter;
  sc_core::sc_port< i_long_long_receiver  , 1 , sc_core::SC_ONE_OR_MORE_BOUND > p1;
  sc_core::sc_port< i_long_long_sender  , 1 , sc_core::SC_ONE_OR_MORE_BOUND > p2;
  long long d;
  

  inline long long decode(long long x)
{
    int i;
    for (i = 0; i < 10000000; i++) {
      (this) -> d *= x;
    }
    return (this) -> d;
  }
  

  inline void main()
{
    long long inFrame;
    long long outFrame;
    (this) -> d = 1;
    while(1){
      ((this) -> p1) ->  receive (&inFrame);
      outFrame = (this) ->  decode (inFrame);
      (this) ->  wait (33330,sc_core::SC_US,7);
      ((this) -> p2) ->  send (outFrame);
    }
  }
  typedef struct VideoCodec SC_CURRENT_USER_MODULE;
  

  inline VideoCodec(class sc_module_name )
{
    sc_module::m_instance_id = _OoO_Instance_Counter++;
{
      class sc_process_handle main_handle = sc_get_curr_simcontext() ->  create_thread_process ("main", false, static_cast<sc_core::SC_ENTRY_FUNC>(&VideoCodec::main), (this), 0, 3, sc_module::m_instance_id);
      (this) -> sc_module::sensitive << main_handle;
      (this) -> sc_module::sensitive_pos << main_handle;
      (this) -> sc_module::sensitive_neg << main_handle;
    }
    ;
  }
}
;
unsigned int VideoCodec::_OoO_Instance_Counter = 0;

struct AudioCodec : public sc_module
{
  static unsigned int _OoO_Instance_Counter;
  sc_core::sc_port< i_long_long_receiver  , 1 , sc_core::SC_ONE_OR_MORE_BOUND > p1;
  sc_core::sc_port< i_long_long_sender  , 1 , sc_core::SC_ONE_OR_MORE_BOUND > p2;
  long long d;
  

  inline long long decode(long long x)
{
    int i;
    for (i = 0; i < 7500000; i++) {
      (this) -> d *= x;
    }
    return (this) -> d;
  }
  

  inline void main()
{
    long long inFrame;
    long long outFrame;
    (this) -> d = 1;
    while(1){
      ((this) -> p1) ->  receive (&inFrame);
      outFrame = (this) ->  decode (inFrame);
      (this) ->  wait (26120,sc_core::SC_US,12);
      ((this) -> p2) ->  send (outFrame);
    }
  }
  typedef struct AudioCodec SC_CURRENT_USER_MODULE;
  

  inline AudioCodec(class sc_module_name )
{
    sc_module::m_instance_id = _OoO_Instance_Counter++;
{
      class sc_process_handle main_handle = sc_get_curr_simcontext() ->  create_thread_process ("main", false, static_cast<sc_core::SC_ENTRY_FUNC>(&AudioCodec::main), (this), 0, 8, sc_module::m_instance_id);
      (this) -> sc_module::sensitive << main_handle;
      (this) -> sc_module::sensitive_pos << main_handle;
      (this) -> sc_module::sensitive_neg << main_handle;
    }
    ;
  }
}
;
unsigned int AudioCodec::_OoO_Instance_Counter = 0;

struct DUT : public sc_module
{
  static unsigned int _OoO_Instance_Counter;
  sc_core::sc_port< i_long_long_receiver  , 1 , sc_core::SC_ONE_OR_MORE_BOUND > VideoIn;
  sc_core::sc_port< i_long_long_receiver  , 1 , sc_core::SC_ONE_OR_MORE_BOUND > AudioInL;
  sc_core::sc_port< i_long_long_receiver  , 1 , sc_core::SC_ONE_OR_MORE_BOUND > AudioInR;
  sc_core::sc_port< i_long_long_sender  , 1 , sc_core::SC_ONE_OR_MORE_BOUND > VideoOut;
  sc_core::sc_port< i_long_long_sender  , 1 , sc_core::SC_ONE_OR_MORE_BOUND > AudioOutLeft;
  sc_core::sc_port< i_long_long_sender  , 1 , sc_core::SC_ONE_OR_MORE_BOUND > AudioOutRight;
  struct VideoCodec video;
  struct AudioCodec left;
  struct AudioCodec right;
  

  virtual inline void before_end_of_elaboration()
{
    (this) -> video . VideoCodec::p1 . bind(((this) -> VideoIn));
    (this) -> video . VideoCodec::p2 . bind(((this) -> VideoOut));
    (this) -> left . AudioCodec::p1 . bind(((this) -> AudioInL));
    (this) -> left . AudioCodec::p2 . bind(((this) -> AudioOutLeft));
    (this) -> right . AudioCodec::p1 . bind(((this) -> AudioInR));
    (this) -> right . AudioCodec::p2 . bind(((this) -> AudioOutRight));
  }
  typedef struct DUT SC_CURRENT_USER_MODULE;
  

  inline DUT(class sc_module_name ) : video(("video")), left(("left")), right(("right"))
{
  }
}
;
unsigned int DUT::_OoO_Instance_Counter = 0;

struct Display : public sc_module
{
  static unsigned int _OoO_Instance_Counter;
  sc_core::sc_port< i_long_long_receiver  , 1 , sc_core::SC_ONE_OR_MORE_BOUND > VideoIn;
/* Frame */
  

  inline void display(long long )
{
#ifdef PRINT_FRAMES
// display the frame
    putchar('V');
    fflush(stdout);
#endif
  }
  

  inline void main()
{
    long long Frame;
    while(1){
      ((this) -> VideoIn) ->  receive (&Frame);
      (this) ->  display (Frame);
    }
  }
  typedef struct Display SC_CURRENT_USER_MODULE;
  

  inline Display(class sc_module_name )
{
    sc_module::m_instance_id = _OoO_Instance_Counter++;
{
      class sc_process_handle main_handle = sc_get_curr_simcontext() ->  create_thread_process ("main", false, static_cast<sc_core::SC_ENTRY_FUNC>(&Display::main), (this), 0, 13, sc_module::m_instance_id);
      (this) -> sc_module::sensitive << main_handle;
      (this) -> sc_module::sensitive_pos << main_handle;
      (this) -> sc_module::sensitive_neg << main_handle;
    }
    ;
  }
}
;
unsigned int Display::_OoO_Instance_Counter = 0;

struct Speaker : public sc_module
{
  static unsigned int _OoO_Instance_Counter;
  sc_core::sc_port< i_long_long_receiver  , 1 , sc_core::SC_ONE_OR_MORE_BOUND > AudioIn;
  char Channel;
/* Frame */
  

  inline void sound(long long )
{
#ifdef PRINT_FRAMES
// output the sound
    putchar(((this) -> Channel));
    fflush(stdout);
#endif
  }
  

  inline void main()
{
    long long Frame;
    while(1){
      ((this) -> AudioIn) ->  receive (&Frame);
      (this) ->  sound (Frame);
    }
  }
  

  inline Speaker(class sc_module_name name,char chn) : sc_module(name), Channel(chn)
{
    sc_module::m_instance_id = _OoO_Instance_Counter++;
{
      class sc_process_handle main_handle = sc_get_curr_simcontext() ->  create_thread_process ("main", false, static_cast<sc_core::SC_ENTRY_FUNC>(&Speaker::main), (this), 0, 16, sc_module::m_instance_id);
      (this) -> sc_module::sensitive << main_handle;
      (this) -> sc_module::sensitive_pos << main_handle;
      (this) -> sc_module::sensitive_neg << main_handle;
    }
    ;
  }
  typedef struct Speaker SC_CURRENT_USER_MODULE;
}
;
unsigned int Speaker::_OoO_Instance_Counter = 0;

struct Monitor : public sc_module
{
  static unsigned int _OoO_Instance_Counter;
  sc_core::sc_port< i_long_long_receiver  , 1 , sc_core::SC_ONE_OR_MORE_BOUND > Video;
  sc_core::sc_port< i_long_long_receiver  , 1 , sc_core::SC_ONE_OR_MORE_BOUND > AudioLeft;
  sc_core::sc_port< i_long_long_receiver  , 1 , sc_core::SC_ONE_OR_MORE_BOUND > AudioRight;
  struct Display screen;
  struct Speaker speakerL;
  struct Speaker speakerR;
  

  virtual inline void before_end_of_elaboration()
{
    (this) -> screen . Display::VideoIn . bind(((this) -> Video));
    (this) -> speakerL . Speaker::AudioIn . bind(((this) -> AudioLeft));
    (this) -> speakerR . Speaker::AudioIn . bind(((this) -> AudioRight));
  }
  typedef struct Monitor SC_CURRENT_USER_MODULE;
  

  inline Monitor(class sc_module_name ) : screen(("screen")), speakerL(("speakerL"),'L'), speakerR(("speakerR"),'R')
{
  }
}
;
unsigned int Monitor::_OoO_Instance_Counter = 0;

struct Top : public sc_module
{
  static unsigned int _OoO_Instance_Counter;
  class c_long_long_queue VideoStream;
  class c_long_long_queue AudioStreamL;
  class c_long_long_queue AudioStreamR;
  class c_long_long_queue Video;
  class c_long_long_queue AudioLeft;
  class c_long_long_queue AudioRight;
  struct Stimulus stimulus;
  struct DUT dut;
  struct Monitor monitor;
  

  virtual inline void before_end_of_elaboration()
{
    (this) -> stimulus . Stimulus::VideoStream . bind(((this) -> VideoStream));
    (this) -> stimulus . Stimulus::AudioStreamL . bind(((this) -> AudioStreamL));
    (this) -> stimulus . Stimulus::AudioStreamR . bind(((this) -> AudioStreamR));
    (this) -> dut . DUT::VideoIn . bind(((this) -> VideoStream));
    (this) -> dut . DUT::AudioInL . bind(((this) -> AudioStreamL));
    (this) -> dut . DUT::AudioInR . bind(((this) -> AudioStreamR));
    (this) -> dut . DUT::VideoOut . bind(((this) -> Video));
    (this) -> dut . DUT::AudioOutLeft . bind(((this) -> AudioLeft));
    (this) -> dut . DUT::AudioOutRight . bind(((this) -> AudioRight));
    (this) -> monitor . Monitor::Video . bind(((this) -> Video));
    (this) -> monitor . Monitor::AudioLeft . bind(((this) -> AudioLeft));
    (this) -> monitor . Monitor::AudioRight . bind(((this) -> AudioRight));
  }
  typedef struct Top SC_CURRENT_USER_MODULE;
  

  inline Top(class sc_module_name ) : VideoStream(("VideoStream"),10ul), AudioStreamL(("AudioStreamL"),10ul), AudioStreamR(("AudioStreamR"),10ul), Video(("Video"),10ul), AudioLeft(("AudioLeft"),10ul), AudioRight(("AudioRight"),10ul), stimulus(("stimulus")), dut(("dut")), monitor(("monitor"))
{
  }
}
;
unsigned int Top::_OoO_Instance_Counter = 0;
bool sc_core::_OoO_Data_Conflict_Table[100] = {
1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 
1, 1, 1, 0, 0, 1, 0, 0, 0, 0, 
1, 1, 1, 0, 0, 1, 0, 0, 0, 0, 
1, 0, 0, 1, 1, 0, 1, 0, 0, 0, 
1, 0, 0, 1, 1, 0, 1, 0, 0, 0, 
1, 1, 1, 0, 0, 1, 0, 0, 0, 0, 
1, 0, 0, 1, 1, 0, 1, 0, 0, 0, 
1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 
1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 
1, 0, 0, 0, 0, 0, 0, 1, 1, 1};
unsigned int sc_core::_OoO_Data_Conflict_Table_Size = 10;
bool sc_core::_OoO_Event_Notify_Table[100] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
unsigned int sc_core::_OoO_Event_Notify_Table_Size = 10;
int sc_core::_OoO_Conflict_Index_Lookup_Table[49] = {0, 3, 7, 8, 12, 13, 16, 0, 1, 2, 3, 4, 7, 8, 0, -1, -1, 5, 6, -1, 9, 0, -1, -1, -1, -1, -1, -1, 0, -1, -1, -1, -1, -1, -1, 0, -1, -1, -1, -1, -1, -1, 0, -1, -1, -1, -1, -1, -1};
unsigned int sc_core::_OoO_Max_Number_of_Instances = 6;
unsigned int sc_core::_OoO_Number_of_Segments = 7;
long long sc_core::_OoO_Curr_Time_Advance_Table_Time[7] = {0LL, 0LL, 33330000000000LL, 0LL, 26120000000000LL, 0LL, 0LL};
int sc_core::_OoO_Curr_Time_Advance_Table_Delta[7] = {0, 0, 0, 0, 0, 0, 0};
unsigned int sc_core::_OoO_Curr_Time_Advance_Table_Size = 7;
long long sc_core::_OoO_Next_Time_Advance_Table_Time[7] = {-1LL, 33330000000000LL, 33330000000000LL, 26120000000000LL, 26120000000000LL, -1LL, -1LL};
int sc_core::_OoO_Next_Time_Advance_Table_Delta[7] = {-1, 0, 0, 0, 0, -1, -1};
unsigned int sc_core::_OoO_Next_Time_Advance_Table_Size = 7;
int sc_core::_OoO_Time_Advance_Index_Lookup_Table[7] = {0, 3, 7, 8, 12, 13, 16};
unsigned int sc_core::_OoO_Time_Advance_Index_Lookup_Table_Size = 7;

int sc_main(int argc,char *argv[])
{
  struct Top top(("top"));
  printf("Playing...\n");
  sc_start();
  return 0;
}
