#ifndef CONFLICT_TABLE_H_INCLUDED_
#define CONFLICT_TABLE_H_INCLUDED_

#include<utility>

#include "../internal_representation/path_instance_mapper.h"
#include "segment_graph.h"

namespace risc {

namespace sg {

class ConflictTable {
public:

  /**
   * \brief Default constructor
   */
  ConflictTable(
      SegmentGraph &segment_graph,
      PathInstanceMapper &path_instance_mapper,
      bool debugging = false);

  /**
   * \brief Default constructor
   */
  ~ConflictTable();

  /**
   * \brief This lookup table translates segment ids into vertex descriptors.
   *        At position i is the segment with the id of
   *        instance_id_to_index_lookup_[0][i]
   */
  VertexDescriptor *vertex_lookup_;

  /**
   * \brief This variable defines the size of the conflict table.
   *        The size is quadratic.
   */
  int size_of_conflict_table_;

  /**
   * \brief This variable represents the max number of module instances in the
   *        design. Also, this is the number of rows in the
   *        instance_id_to_index_lookup_ table.
   */
  int max_instances_;

  /**
   * \brief This variable defines how many segments exist in the segment
   *        graph. Also, this is the number of columns in the
   *        instance_id_to_index_lookup_ table.
   */
  int number_of_segments_;

  /**
   * \brief This table should be used a lookup table:
   *        (segment id, instance id) -> index of conflict table
   */
  int **instance_id_to_index_lookup_;

  /**
   * \brief The conflict table
   */
  bool **conflict_table_;

  /**
   * \function print_conflict_table
   * \brief    Generates a html file of the conflict table
   */
  void print_conflict_table(std::string filename);

  /**
   * \function print_lookup_table
   * \brief    Generates a html file of the lookup table
   */
  void print_lookup_table(std::string filename);

  /**
   * \function segment_and_instance_id_to_index
   * \brief    This function translates a given segment id and instance id
   *           to a index of the conflict table
   */
  int segment_and_instance_id_to_index(int segment_id, int instance_id);

  /**
   * \function index_to_segment_and_instance_id
   * \brief    This function translates a given index from the conflict table
   *           into the corresponding segment id and instance id. The first
   *           value is the segment id and the second value is the instance id.
   */
  std::pair<int, int> index_to_segment_and_instance_id(int index);

  /**
   * \function has_conflict
   * \brief    This function looks up if there is a conflict between segment
   *           1 with instance id 1 and segment 2 with instance id 2
   */
  bool has_conflict(int segment_id_1, int instance_id_1, int segment_id_2,
      int instance_id_2);

  /**
   * \function print_instance_id_to_index_lookup_table
   * \brief    This function prints the table instance_id_to_index_lookup_table
   *           in the terminal
   */
  void print_instance_id_to_index_lookup_table();

  /**
   * \function set_conflict
   * \brief    This function sets a conflict in the conflict table
   */
  void set_conflict(int segment_id_1, int instance_id_1,
      int segment_id_2, int instance_id_2);

  /**
   * \function set_conflict
   * \brief    This function sets a conflict in the conflict table
   */
  void set_conflict_cause(int segment_id_1, int instance_id_1,
      int segment_id_2, int instance_id_2, SgNode *cause = NULL);

  /**
   * \function determine_conflict_table
   * \brief    This function determines the conflict table.
   *           The function is overloaded for the data and event conflict table.
   */
  virtual void determine_conflict_table();


  /**
   * \function segment_id_to_vertex_descriptor
   * \brief    This function translates the given segment id to the
   *           corresponding vertex descriptor
   */
  VertexDescriptor segment_id_to_vertex_descriptor(int id);

  PathInstanceMapper &path_instance_mapper_;
  SegmentGraph &segment_graph_;

  // The following part is for debugging purposes
  struct ConflictInformation {
    std::set<SgNode*> conflict_variables_;
  };

  ConflictInformation **conflict_info_;

  bool debugging_;
};

}; // end of namesapce sg

}; // end of namespace risc

#endif /* CONFLICT_TABLE_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
